@@ -1,7 +1,12 @@
 """
-Simple Example of using Pydantic to generate a CLI
+# Simple Example of using Pydantic to generate a CLI
 
-This example will generate a CLI tool with 2 (required) positional arguments
+This example will generate a CLI tool with 2 **required** arguments
+and can be called
+
+```bash
+my-tool --input_file file.fasta --max_records 10
+```
 """
 from pydantic import BaseModel
 
@@ -19,4 +24,4 @@ def example_runner(opts: Options) -> int:
 
 
 if __name__ == "__main__":
-    run_and_exit(Options, example_runner, description=__doc__, version='0.1.0')
+    run_and_exit(Options, example_runner, description=__doc__, version="0.1.0")
